/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Sat May 18 15:23:21 CEST 2002: included cpu autodetection from Magnus Ekdahl
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

#include "readdb.h"
#include "defaults.h"
#include "others.h"
#include "options.h"
#include "manager.h"
#include "treewalk.h"
#include "shared.h"

int scanmanager(const struct optstruct *opt)
{
	mode_t fmode;
	sigset_t signal_set;
	int ret = 0, threads = 0, compression = 0, fmodeint;
	struct node *trie;
	struct passwd *user = NULL;
	char *fullpath = NULL, cwd[200];

#ifndef C_CYGWIN
    if(!getuid()) {
	if((user = getpwnam(UNPUSER)) == NULL) {
	    mprintf("@Can't get information about user "UNPUSER".\n");
	    exit(60); /* this is critical problem, so we just exit here */
	}
    }
#endif

    if(optl(opt, "threads")) {
	if(threads < 0 || !sisnumber(getargl(opt, "threads"))) {
	    mprintf("@Invalid threads number.\n");
	    return 51;
	}
	threads = atoi(getargl(opt, "threads"));
    }

#ifdef C_LINUX
    else /* Autodetecting threads */
      {
	if((threads = detectCpu()) > 1)
	    mprintf("Autodetected %d CPUs. Starting %d threads.\n", threads, threads);
	else
	    threads = 0;
      }
#endif

    if(optl(opt, "unzip") || optl(opt, "unrar") || optl(opt, "unace") ||
       optl(opt, "unarj") || optl(opt, "zoo") || optl(opt, "jar") ||
       optl(opt, "lha") || optl(opt, "tar") || optl(opt, "tgz") ||
       optl(opt, "deb"))
	    compression = 1;


    /* now initialize the database from hardcoded file or --database */

    if(optc(opt, 'd'))
	trie = readdb(getargc(opt, 'd'), NULL);
    else {
	    char *pt;

	trie = readdb(VIRUSDB, NULL);
	/* when '-d' not specified, load all .db files from DATADIR also */
	while((pt = listdb(DATADIR))) {
	    trie = readdb(pt, trie);
	    free(pt);
	}
    }

    if(!trie) {
	mprintf("@Can't initialize virus database.\n");
	return 50;
    }

    /* build the proper trie */
    failtrans(trie);
    fasttrie(trie, trie);

    if(compression) {
	/* block all signals */

	sigfillset(&signal_set); /* all signals */
	sigprocmask(SIG_SETMASK, &signal_set, NULL); /* set lock */
    }

    /* check filetype */
    if(!strcmp(opt->filename, "-")) { /* read data from stdin */
	ret = checkstdin(trie);

    } else if(strlen(opt->filename) == 0) {

	/* we need full path for some reasons (eg. archive handling) */
	if(!getcwd(cwd, 200)) {
	    mprintf("@Can't get absolute pathname of current working directory.\n");
	    ret = 57;
	} else
	    ret = scandirs(cwd, trie, threads, user, opt);

    } else if((fmodeint = info(opt->filename, 2)) == -1) {
	mprintf("@Can't access file %s\n", opt->filename);
	perror(opt->filename);
	ret = 56;
    } else {
	fmode = (mode_t) fmodeint;

	if(compression && (opt->filename[0] != '/')) {
	    /* we need to complete the path */
	    if(!getcwd(cwd, 200)) {
		mprintf("@Can't get absolute pathname of current working directory.\n");
		return 57;
	    } else {
		fullpath = mcalloc(512, sizeof(char));
		snprintf(fullpath, 512, "%s/%s", cwd, opt->filename);
		mprintf("*Full path: %s\n", fullpath);
	    }
	} else 
	    fullpath = (char *) opt->filename;

	switch(fmode & S_IFMT) {
	    case S_IFREG: 
		ret = scanfile(fullpath, trie, threads, user, opt);
		break;

	    case S_IFDIR:
		  ret = scandirs(fullpath, trie, threads, user, opt);
		  break;
	    default:
		  mprintf("@Not supported file type (%s)\n", opt->filename);
		  ret = 52;
	}
    }

    if(compression)
	free(fullpath);

    if(ret < 50) { /* hopefully no error detected */
	/* overwrite return code */

	if(claminfo.ifiles)
	    ret = 1;
	else
	    ret = 0;
    }

    return ret;
}

int scanfile(const char *filename, struct node *root, int thread, const struct passwd *user, const struct optstruct *opt)
{
	int ret;

    if(info(filename, 1) == 0) {
	if(!printinfected)
	    mprintf("%s: Empty file.\n", filename);
	return 0;
    }

    /* check the extension */

    if((strbcasestr(filename, ".zip") && optl(opt, "unzip")) 
    || (strbcasestr(filename, ".rar") && optl(opt, "unrar"))
    || (strbcasestr(filename, ".ace") && optl(opt, "unace"))
    || (strbcasestr(filename, ".arj") && optl(opt, "unarj"))
    || (strbcasestr(filename, ".zoo") && optl(opt, "zoo"))
    || (strbcasestr(filename, ".jar") && optl(opt, "jar"))
    || (strbcasestr(filename, ".lzh") && optl(opt, "lha"))
    || (strbcasestr(filename, ".tar") && optl(opt, "tar"))
    || (strbcasestr(filename, ".deb") && optl(opt, "deb"))
    || ( (strbcasestr(filename, ".tar.gz") || strbcasestr(filename, ".tgz")) 
	 && (optl(opt, "tgz") || optl(opt, "deb"))  )) {

	/* check permissions */
	switch(readaccess(filename, UNPUSER)) {
	    case -1:
		mprintf("@Can't get information about user "UNPUSER".\n");
		exit(60); /* this is critical problem, so we just exit here */
	    case -2:
		mprintf("@Can't get information about current user.\n");
		exit(59); /* this is critical problem, so we just exit here */
	    case 0: /* read access denied */
		if(getuid()) {
		    if(!printinfected)
			mprintf("%s: Access denied to archive.\n", filename);
		} else {
			int maxspace;

		    if(optl(opt, "max-space")) {
			    char *cpy, *ptr;
			ptr = getargl(opt, "max-space");
			if(tolower(ptr[strlen(ptr) - 1]) == 'm') {
			    cpy = mcalloc(strlen(ptr), sizeof(char));
			    strncpy(cpy, ptr, strlen(ptr) - 1);
			    maxspace = atoi(cpy) * 1024;
			    free(cpy);
			} else /* default - kilobytes */
			    maxspace = atoi(ptr);
		    } else
			maxspace = 0;

		    if(maxspace)
			if(info(filename, 1) / 1024 > maxspace) {
			    if(!printinfected)
				mprintf("%s: Archive too big.\n", filename);
			    return 0;
			}

		    return(scandenied(filename, root, thread, user, opt));
		}
		return 0;
	    case 1:
		return(scancompressed(filename, root, thread, user, opt));
	}
    }

    if(getuid())
	switch(readaccess(filename, NULL)) {
	    case -2:
		mprintf("@Can't get information about current user.\n");
		exit(59); /* this is critical problem, so we just exit here */
	    case 0: /* read access denied */
		if(!printinfected)
		    mprintf("%s: Access denied.\n", filename);
		return 0;
	}

    if((ret = checkfile(filename, root))) {
	if(optl(opt, "remove")) {
	    if(unlink(filename)) {
		mprintf("%s: Can't remove.\n", filename);
		logg("%s: Can't remove.\n", filename);
		claminfo.notremoved++;
	    } else {
		mprintf("%s: Removed.\n", filename);
		logg("%s: Removed.\n", filename);
	    }
	}
    }
    return ret;
}

/* it has guaranteed read access to the archive */
int scancompressed(const char *filename, struct node *root, int thread, const struct passwd *user, const struct optstruct *opt)
{
	int ret = 0;
	char *tmpdir, *userprg;
	struct stat statbuf;

    stat(filename, &statbuf);

    if(!S_ISREG(statbuf.st_mode)) {
	mprintf("^Suspected archive %s is not a regular file.\n", filename);
	return 0;
    }

    /* check write access */

    if((tmpdir = getargl(opt, "tempdir")) == NULL)
	tmpdir = "/tmp";

    if(writeaccess(tmpdir, UNPUSER) != 1) {
	mprintf("@Can't write to the temporary directory.\n");
	exit(64);
    }

    /* generate the temporary directory */

    tmpdir = gentemp(getargl(opt, "tempdir"));
    if(mkdir(tmpdir, 0700)) {
	mprintf("@Can't create the temporary directory %s\n", tmpdir);
	exit(63); /* critical */
    }

    if(user)
	chown(tmpdir, user->pw_uid, user->pw_gid);


    /* unpack file  - as unprivileged user */
    if(strbcasestr(filename, ".zip")) {
	char *args[] = { "unzip", "-P", "clam", "-o", (char *) filename, NULL };

	if((userprg = getargl(opt, "unzip")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unzip", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".rar")) { 
	char *args[] = { "unrar", "x", "-p-", "-y", (char *) filename, NULL };
	if((userprg = getargl(opt, "unrar")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unrar", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".ace")) { 
	char *args[] = { "unace", "x", "-y", (char *) filename, NULL };
	if((userprg = getargl(opt, "unace")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unace", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".arj")) { 
	char *args[] = { "unarj", "x", (char *) filename, NULL };
	if((userprg = getargl(opt, "unarj")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unarj", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".zoo")) { 
	char *args[] = { "zoo", "x.O", (char *) filename, NULL };
	if((userprg = getargl(opt, "zoo")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("zoo", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".jar")) { 
	char *args[] = { "unzip", "-P", "clam", "-o", (char *) filename, NULL };
	if((userprg = getargl(opt, "jar")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unzip", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".lzh")) { 
	char *args[] = { "lha", "xf", (char *) filename, NULL };
	if((userprg = getargl(opt, "lha")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("lha", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".tar")) { 
	char *args[] = { "tar", "-xpvf", (char *) filename, NULL };
	if((userprg = getargl(opt, "tar")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("tar", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".deb")) { 
	char *args[] = { "ar", "x", (char *) filename, NULL };
	if((userprg = getargl(opt, "deb")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("ar", args, tmpdir, user, opt);

    } else if((strbcasestr(filename, ".tar.gz") || strbcasestr(filename, ".tgz"))) {
	char *args[] = { "tar", "-zxpvf", (char *) filename, NULL };
	if((userprg = getargl(opt, "tgz")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("tar", args, tmpdir, user, opt);
    }

    /* fix permissions of extracted files */
    fixperms(tmpdir);

    if(!ret) /* execute successful */
	/* notice, that we can't use threadscan() so also scandirs() here */
	ret = treewalk(tmpdir, root, thread, user, opt);

    /* remove the directory  - as clamav */
    clamav_rmdirs(tmpdir);

    /* free tmpdir - it's not necessary now */
    free(tmpdir);

    switch(ret) {
	case -1:
	    mprintf("@Can't fork().\n");
	    exit(61); /* this is critical problem, so we just exit here */
	case -2:
	    mprintf("@Can't execute some unpacker. Check paths and permissions on the temporary directory.\n");
	    exit(62); /* this is critical problem, so we just exit here */
	case -3:
	    return 0;
	case 0:
	    /* no viruses found in archive, we scan just in case the same
	     * archive
	     */
	    if(!printinfected)
		mprintf("(raw) ");

	    if((ret = checkfile(filename, root))) {
		if(optl(opt, "remove")) {
		    if(unlink(filename)) {
			mprintf("%s: Can't remove.\n", filename);
			logg("%s: Can't remove.\n", filename);
			claminfo.notremoved++;
		    } else {
			mprintf("%s: Removed.\n", filename);
			logg("%s: Removed.\n", filename);
		    }
		}
	    }
	    return ret;
	case 1:
	    logg("%s: Infected Archive FOUND\n", filename);
	    mprintf("%s: Infected Archive FOUND\n", filename);
	    if(optl(opt, "remove")) {
		if(unlink(filename)) {
		    mprintf("%s: Can't remove.\n", filename);
		    logg("%s: Can't remove.\n", filename);
		    claminfo.notremoved++;
		} else {
		    mprintf("%s: Removed.\n", filename);
		    logg("%s: Removed.\n", filename);
		}
	    }
	    return 1;
	default:
	    mprintf("@Strange value (%d) returned in scancompressed()\n", ret);
	    return 0;
    }
}

int scandenied(const char *filename, struct node *root, int thread, const struct passwd *user, const struct optstruct *opt)
{
	char *tmpdir, *tmpfile, *pt;
	struct stat statbuf;
	int ret;

    stat(filename, &statbuf);
    if(!S_ISREG(statbuf.st_mode)) {
	mprintf("^Suspected archive %s is not a regular file.\n", filename);
	return 0;
    }

    /* check write access */

    if((tmpdir = getargl(opt, "tempdir")) == NULL)
	tmpdir = "/tmp";

    if(writeaccess(tmpdir, UNPUSER) != 1) {
	mprintf("@Can't write to the temporary directory.\n");
	exit(64);
    }

    tmpdir = gentemp(getargl(opt, "tempdir"));
    if(mkdir(tmpdir, 0700)) {
	mprintf("@Can't create the temporary directory %s\n", tmpdir);
	exit(63); /* critical */
    }

    tmpfile = (char *) mcalloc(strlen(tmpdir) + strlen(filename) + 10, sizeof(char));
    pt = strrchr(filename, '/');
    if(!pt)
	pt = (char *) filename;
    else
	pt += 1;

    sprintf(tmpfile, "%s/%s", tmpdir, pt);

    if(copyfile(filename, tmpfile) == -1) {
	mprintf("!I/O error.\n");
	perror("copyfile()");
	exit(58);
    }

    fixperms(tmpdir);

    if(user) {
	chown(tmpdir, user->pw_uid, user->pw_gid);
	chown(tmpfile, user->pw_uid, user->pw_gid);
    }

    if((ret = treewalk(tmpdir, root, thread, user, opt)) == 1) {
	logg("(Real infected archive: %s)\n", filename);
	mprintf("(Real infected archive: %s)\n", filename);

	if(optl(opt, "remove")) {
	    if(unlink(filename)) {
		mprintf("%s: Can't remove.\n", filename);
		logg("%s: Can't remove.\n", filename);
		claminfo.notremoved++;
	    } else {
	        mprintf("%s: Removed.\n", filename);
	        logg("%s: Removed.\n", filename);
	    }
        }
    }

    /* remove the directory  - as clamav */
    clamav_rmdirs(tmpdir);

    free(tmpdir);
    free(tmpfile);

    return ret;
}

int scandirs(const char *dirname, struct node *root, int thread, const struct passwd *user, const struct optstruct *opt)
{

#ifdef THREADS
    if(thread)
	return(threadscan(dirname, root, thread, user, opt));
    else
#endif
	return(treewalk(dirname, root, thread, user, opt));
}
