/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#ifdef THREADS
#  include <pthread.h>
#  include <semaphore.h>
#endif

#include "shared.h"
#include "readdb.h"
#include "manager.h"
#include "others.h"
#include "options.h"
#include "treewalk.h"

/*
 * recursive functions, similar to ftw(), these functions are duplicated below
 * with some small changes.
 */

int treewalk(const char *dirname, struct node *root, int thread, const struct passwd *user, const struct optstruct *opt)
{
	DIR *dd;
	struct dirent *dent;
	struct stat statbuf;
	char fname[256];
	int scanret = 0;

    claminfo.dirs++;

    if((dd = opendir(dirname)) != NULL) {
	while((dent = readdir(dd))) {
	    if(dent->d_ino) {
		if(strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..")) {
		    /* build the full name */
		    if(strlen(dirname) + strlen(dent->d_name) + 2 <= 256)
			snprintf(fname, 256, "%s/%s", dirname, dent->d_name);
		    else continue;

		    /* stat the file */
		    if(lstat(fname, &statbuf) != -1) {
			if(S_ISDIR(statbuf.st_mode) && !S_ISLNK(statbuf.st_mode) && recursion)
			    treewalk(fname, root, thread, user, opt);
			else
			    if(S_ISREG(statbuf.st_mode))
				scanret += scanfile(fname, root, thread, user, opt);
		    }
		}
	    }
	}
    } else {
	mprintf("@Can't open directory %s\n", dirname);
	return 53;
    }

    closedir(dd);

    if(scanret)
	return 1;
    else
	return 0;

}

int rmdirs(const char *dirname)
{
	DIR *dd;
	struct dirent *dent;
	struct stat maind, statbuf;
	char fname[256];

    if((dd = opendir(dirname)) != NULL) {
	while(stat(dirname, &maind) != -1) {
	    if(!rmdir(dirname)) break;

	    while((dent = readdir(dd))) {
		if(dent->d_ino) {
		    if(strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..")) {
			/* build the full name */
			if(strlen(dirname) + strlen(dent->d_name) + 2 <= 256)
			    snprintf(fname, 256, "%s/%s", dirname, dent->d_name);
			else continue;

			/* stat the file */
			if(lstat(fname, &statbuf) != -1) {
			    if(S_ISDIR(statbuf.st_mode) && !S_ISLNK(statbuf.st_mode)) {
				if(rmdir(fname) == -1) { /* can't be deleted */
				    if(errno == EACCES) {
					mprintf("@Can't remove some temporary directories due to access problem.\n");
					return 0;
				    }
				    rmdirs(fname);
				}
			    } else
				unlink(fname);
			}
		    }
		}
	    }

	    rewinddir(dd);

	}

    } else { 
	mprintf("@Can't open directory %s\n", dirname);
	return 53;
    }

    closedir(dd);
    return 0;
}

int fixperms(const char *dirname)
{
	DIR *dd;
	struct dirent *dent;
	struct stat statbuf;
	char fname[256];
	int scanret = 0;

    if((dd = opendir(dirname)) != NULL) {
	while((dent = readdir(dd))) {
	    if(dent->d_ino) {
		if(strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..")) {
		    /* build full name */
		    if(strlen(dirname) + strlen(dent->d_name) + 2 <= 256)
			snprintf(fname, 256, "%s/%s", dirname, dent->d_name);
		    else continue;

		    /* stat the file */
		    if(lstat(fname, &statbuf) != -1) {
			if(S_ISDIR(statbuf.st_mode) && !S_ISLNK(statbuf.st_mode)) {
			    chmod(fname, 0700);
			    fixperms(fname);
			} else
			    chmod(fname, 0700);
		    }
		}
	    }
	}
    } else {
	mprintf("@Can't open directory %s\n", dirname);
	return 53;
    }

    closedir(dd);

    if(scanret)
	return 1;
    else
	return 0;

}

int du(const char *dirname, struct s_du *n)
{
	DIR *dd;
	struct dirent *dent;
	struct stat statbuf;
	char fname[256];

    if((dd = opendir(dirname)) != NULL) {
	while((dent = readdir(dd))) {
	    if(dent->d_ino) {
		if(strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..")) {
		    n->files++;

		    /* build the full name */
		    if(strlen(dirname) + strlen(dent->d_name) + 2 <= 256)
			snprintf(fname, 256, "%s/%s", dirname, dent->d_name);
		    else continue;

		    /* stat the file */
		    if(lstat(fname, &statbuf) != -1) {
			if(S_ISDIR(statbuf.st_mode) && !S_ISLNK(statbuf.st_mode)) {
			    du(fname, n);
			} else {
			    n->space += statbuf.st_size / 1024;
			}
		    }
		}
	    }
	}
    } else {
	mprintf("@Can't open directory %s\n", dirname);
	return 53;
    }

    closedir(dd);

    return 0;
}


#ifdef THREADS

struct filedb {
    char fname[256];
    struct filedb *next;
};

struct threadarg {
    struct node *trie;
    const struct passwd *user;
    const struct optstruct *opt;
    int thread;
};

struct filedb *filelist;

#ifdef DEBUG
pthread_mutexattr_t attr;
pthread_mutex_t filelist_mutex;
#else
pthread_mutex_t filelist_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

sem_t filelist_sem;
int stopper = 0, threadvir = 0;

void *threadfn(void *arg)
{
	struct filedb *file;
	struct threadarg *tharg = (struct threadarg *) arg;


    while(1) {

	sem_wait(&filelist_sem);

	if(stopper)
	    return NULL;

#ifdef DEBUG
	if(pthread_mutex_lock(&filelist_mutex)) {
	    mprintf("!Mutex deadlock.\n");
	    exit(90);
	} else
	    mprintf("Mutex correctly locked.\n");
#else
	pthread_mutex_lock(&filelist_mutex);
#endif

	file = filelist;
	filelist = filelist->next;

#ifdef DEBUG
	if(pthread_mutex_unlock(&filelist_mutex)) {
	    mprintf("Can't unlock the mutex.\n");
	    exit(91);
	} else
	    mprintf("Mutex correctly unlocked.\n");
#else
	pthread_mutex_unlock(&filelist_mutex);
#endif

	threadvir += scanfile(file->fname, tharg->trie, tharg->thread, tharg->user, tharg->opt);

	free(file);
    }

    return NULL;
}

void addfile(const char *filename)
{
	struct filedb *file;

    file = (struct filedb*) mcalloc(1, sizeof(struct filedb));
    strncpy(file->fname, filename, 256);

    pthread_mutex_lock(&filelist_mutex);
    file->next = filelist;
    filelist = file;

    sem_post(&filelist_sem);
    pthread_mutex_unlock(&filelist_mutex);
}

int buildtree(const char *dirname, const struct optstruct *opt)
{
	DIR *dd;
	struct dirent *dent;
	struct stat statbuf;
	char fname[256];

    claminfo.dirs++;

    if((dd = opendir(dirname)) != NULL) {
	while((dent = readdir(dd))) {
	    if(dent->d_ino) {
		if(strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..")) {
		    /* build full name */
		    if(strlen(dirname) + strlen(dent->d_name) + 2 <= 256)
			snprintf(fname, 256, "%s/%s", dirname, dent->d_name);
		    else continue;

		    /* stat the file */
		    if(lstat(fname, &statbuf) != -1) {
			if(S_ISDIR(statbuf.st_mode) && !S_ISLNK(statbuf.st_mode) && optc(opt, 'r'))
			    buildtree(fname, opt);
			else
			    if(S_ISREG(statbuf.st_mode))
				addfile(fname);
		    }
		}
	    }
	}
    } else return 53;

    closedir(dd);
    return 0;
}


int threadscan(const char *dirname, struct node *root, int thread, const struct passwd *user, const struct optstruct *opt)
{
	pthread_t *threadtab;
	pthread_attr_t *attrtab;
	struct threadarg *tharg;
	int i;

    tharg = mcalloc(1, sizeof(struct threadarg));
    tharg->trie = root;
    tharg->user = user;
    tharg->opt = opt;
    tharg->thread = thread;

    filelist = NULL;
    sem_init(&filelist_sem, 0, 0);

#ifdef DEBUG
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_setkind_np(&attr, PTHREAD_MUTEX_ERRORCHECK_NP);
    pthread_mutex_init(&filelist_mutex, &attr);
    pthread_mutexattr_destroy(&attr);
#endif

    threadtab = (pthread_t *) mcalloc(thread, sizeof(pthread_t));
    attrtab = (pthread_attr_t *) mcalloc(thread, sizeof(pthread_attr_t));

    for(i = 0; i < thread; i++) {
	claminfo.threads++;
	pthread_attr_init(&attrtab[i]);
	if(!getuid() && pthread_attr_setschedpolicy(&attrtab[i], SCHED_FIFO)) {
	    mprintf("!setschedpolicy: ");
	    perror("");
	}
	pthread_create(&threadtab[i], &attrtab[i], threadfn, tharg);
	pthread_attr_destroy(&attrtab[i]);
    }

    buildtree(dirname, opt);

    while(1) {
	if(filelist == NULL) {
	    /* set the stopper */
	    stopper = 1;

	    for(i = 0; i < thread; i++)
		sem_post(&filelist_sem);

	    for(i = 0; i < thread; i++)
		pthread_join(threadtab[i], NULL);

	    free(threadtab);

	    mprintf("*threadscan() has finished it's work.\n");

	    if(threadvir)
		return 1;
	    else
		return 0;
	}
	usleep(200000);
    }

    /* impossible ? */
    free(threadtab);
    return -1;
}

#endif /* THREADS */
